 /******************************************************************************

       TAPI serial communication dialing example, 5/25/95 by Mark Gamber
                              (requires Windows 95)

   TAPI example demonstrating how to make a call, make a connection and grab
   the serial handle for serial I/O. A dialog box is used for TAPI control
   based on messages passed to it by the TAPI callback procedure. This method
   doesn't use any global variables or PeekMessage() loops, allowing multiple
   calls without re-entrancy concerns.

   This program and all related files are public domain and may be freely
   distributed. By using the program or any included part, the user assumes
   full responsibility for it's proper use and may not hold the author liable
   for any loss or damage. If unable to accept this condition, the program
   and all related parts may not be used and must be destroyed immediately.

******************************************************************************/

#define WIN32_LEAN_AND_MEAN

#include "windows.h"
#include "windowsx.h"
#include "tapi.h"

// ----------------------------------------------------------------------------
// === Function Prototypes ====================================================
// ----------------------------------------------------------------------------

BOOL WINAPI MainDlgProc( HWND, UINT, WPARAM, LPARAM );
BOOL WINAPI TapiDlgProc( HWND, UINT, WPARAM, LPARAM );
BOOL WINAPI ModemDlgProc( HWND, UINT, WPARAM, LPARAM );
VOID CALLBACK TapiCallbackProc( HANDLE, DWORD, DWORD, DWORD, DWORD, DWORD );
BOOL OpenTAPI( HWND, DWORD, LPSTR );

// ----------------------------------------------------------------------------
// === Global Variables =======================================================
// ----------------------------------------------------------------------------

HINSTANCE hInst;                                //  Application instance handle
HLINEAPP hLineApp;                     //  Application TAPI registration handle
DWORD dwDevices;                                     //  Number of TAPI devices

// ----------------------------------------------------------------------------
// === Application Entry Point ================================================
// ----------------------------------------------------------------------------

int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrev, LPSTR lpCmd,
                      int nShow )
{  
   hInst = hInstance;
   DialogBox( hInstance, MAKEINTRESOURCE( 10000 ), NULL, MainDlgProc );
   return( FALSE );
}

// ----------------------------------------------------------------------------
// === Main Window Procedure (dialog box) =====================================
// ----------------------------------------------------------------------------

BOOL WINAPI MainDlgProc( HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam )
{
   switch( msg )
   {
      case WM_INITDIALOG:
         hLineApp = NULL;                        //  Initialize TAPI app handle
         if( lineInitialize( &hLineApp, hInst,          //  Initialize TAPI DLL
                             (LINECALLBACK)TapiCallbackProc, "TTEST1",
                             &dwDevices ) < 0 )
         {                                             //  If an error occurred
            MessageBox( hDlg, "Error in lineInitialize", "Tapi Test",
                        MB_OK | MB_ICONSTOP );    //  Tell user about the error
            EndDialog( hDlg, FALSE );                    //  and end right here
            return( TRUE );
         }
         EnableWindow( GetDlgItem( hDlg, 101 ), FALSE );  //  Disable "Hang Up"
         return( TRUE );


      case WM_COMMAND:
         if( wParam == IDCANCEL )
         {
            lineShutdown( hLineApp );           //  Shutdown DLL before leaving
            EndDialog( hDlg, TRUE );                //  And end this dialog box
            return( TRUE );                 //  which, of course, kills the app
         }
         if( wParam == 100 )                          //  "Dial" button pressed
         {
            char *pszStr;

            pszStr = GlobalAllocPtr( GHND, 64 );        //  Allocate some space
            if( ! pszStr )                         //  to pass the phone number
               break;                               //  to a control dialog box

            if( GetDlgItemText( hDlg, 102, pszStr, 64 ) )    //  Get the string
               CreateDialogParam( hInst, MAKEINTRESOURCE( 10001 ),
                                  hDlg, TapiDlgProc, //   Start the dial dialog
                                  (LPARAM)pszStr );      //  passing the number
            else
            {
               GlobalFreePtr( pszStr );              //  Else, bitch and scream
               MessageBox( hDlg, "Enter a phone number first", "Tapi Test",
                           MB_OK | MB_ICONEXCLAMATION );
            }
            break;
         }
         break;
   }
   return( FALSE );
}                                                      //  End of MainDlgProc()

// ----------------------------------------------------------------------------
// === Dialog box to handle TAPI connection ===================================
// ----------------------------------------------------------------------------

BOOL WINAPI TapiDlgProc( HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam )
 {
   switch( msg )
   {
      case WM_INITDIALOG:                //  Phone number is passed in a string
      {                                                    //  Open a TAPI line
         if( ! OpenTAPI( hDlg, dwDevices, (LPSTR)lParam ) )
         {
            DestroyWindow( hDlg );       //  If it fails, kill this window, too
            return( TRUE );
         }
         GlobalFreePtr( (LPVOID)lParam ); //  Free string passed here by caller
         return( TRUE );                      //  and continue in modeless mode
      }


      case WM_COMMAND:
         if( wParam == IDCANCEL )       //  Destroy this dialog on cancel which
         {                                //  also kills the connection, if any
            HCALL hCall;
            HLINE hLine;
            LPVARSTRING lpVs;
            LPHANDLE lphComm;
            DWORD dwOut;


            hCall = GetProp( hDlg, "HCALL" );     //  These are attached to the
            hLine = GetProp( hDlg, "HLINE" );   //  window by the TAPI callback
                                         //  As is the current connection state
                  //  Before we close the line, if there's a connection, use it
            if( (DWORD)GetProp( hDlg, "STATE" ) == LINECALLSTATE_CONNECTED )
            {
               lpVs = (LPVARSTRING)GlobalAllocPtr( GHND, //  Allocate some info
                                          sizeof(VARSTRING) + 1024 ); //  space
               if( lpVs )
               {
                  lpVs->dwTotalSize = 1024 + sizeof(VARSTRING);
                  if( ! lineGetID( NULL, 0, hCall,     //  Get info on the call
                                   LINECALLSELECT_CALL, lpVs,    //  which also
                                   "comm/datamodem" ) ) //  gets the COM handle
                  {                 //  Point to the COM handle if GetID worked
                     lphComm = (LPHANDLE)( (LPBYTE)lpVs + sizeof(VARSTRING) );
                     if( *lphComm )
                     {               //  And send something out the serial port
                        WriteFile( *lphComm, "\r\n", 2, &dwOut, NULL );
                        CloseHandle( *lphComm );               //  and close it
                     }
                  }
                  GlobalFreePtr( (LPVOID)lpVs );      //  Free the buffer space
               }
            }

            lineDrop( hCall, NULL, 0 );        //  Drop the line and, if it was
            if( (DWORD)GetProp( hDlg, "STATE" ) == LINECALLSTATE_CONNECTED )
               lineDeallocateCall( hCall );  //  connected, deallocate the call
            
            lineClose( hLine );                              //  Close the line

            RemoveProp( hDlg, "STATE" );       //  Free all the properties that
            RemoveProp( hDlg, "DEVID" ); //  have been attached by the callback
            RemoveProp( hDlg, "HCALL" );
            RemoveProp( hDlg, "HLINE" );
            DestroyWindow( hDlg );                //  Kill this window and exit
            return( TRUE );
         }
         if( wParam == 10000 )       //  Messages posted here from the callback
         {
            char szStr[ 64 ];                                                            
            int i, iItem;

                                              //  Update state if not connected
            if( (DWORD)GetProp( hDlg, "STATE" ) != LINECALLSTATE_CONNECTED )
               SetProp( hDlg, "STATE", (HANDLE)lParam );

            if( lParam == LINECALLSTATE_DISCONNECTED ) //  If line disconnected
               PostMessage( hDlg, WM_COMMAND, IDCANCEL, 0 );      //  handle it

            for( i = 0; i < 16; i++ )      //  Move through possible state bits
            {
               if( ( lParam >> i ) & 1 )                //  If state bit is set
               {                            //  Load string pertaining to state
                  LoadString( hInst, 100 + i, szStr, 64 );
                  iItem = SendDlgItemMessage( hDlg, 100, LB_ADDSTRING, 0,
                                              (LPARAM)szStr );
                  if( iItem != LB_ERR )          //  and add to display listbox
                     SendDlgItemMessage( hDlg, 100, LB_SETCURSEL, iItem, 0 );
               }
            }
            break;
         }
         break;
   }
   return( FALSE );
}                                                      //  End of TapiDlgProc()

// ----------------------------------------------------------------------------
// === TAPI Connection Callback ===============================================
// ----------------------------------------------------------------------------

VOID CALLBACK TapiCallbackProc( HANDLE hDevice, DWORD dwMsg, DWORD dwInst,
                          DWORD dwParam1, DWORD dwParam2, DWORD dwParam3 )
{
   LINECALLINFO lci;


   if( dwMsg == LINE_CALLSTATE )      //  Pass useful information to dialog box
   {                               //  associated to thie instance via lineOpen
      memset( &lci, 0, sizeof(LINECALLINFO) );         //  Initialize structure
      lci.dwTotalSize = sizeof(LINECALLINFO);        //  Specify structure size
      if( lineGetCallInfo( hDevice, &lci ) < 0 )  //  Return on error condition
         return;

      if( dwInst )                     //  This should be a valid window handle
      {                                                     //  and if it is...
         SetProp( (HWND)dwInst, "HCALL", hDevice );      //  Save the hCall and
         SetProp( (HWND)dwInst, "HLINE", (HANDLE)lci.hLine ); //  hLine handles
         PostMessage( (HWND)dwInst, WM_COMMAND, 10000, 
                      dwParam1 );                //  and tell it something's up
      }
   }
   return;
}                                                 //  Short TAPI callback, huh?

// ----------------------------------------------------------------------------
// === Open and Initialize TAPI line ==========================================
// ----------------------------------------------------------------------------

BOOL OpenTAPI( HWND hWnd, DWORD dwNum, LPSTR lpszPhoneNumber )
{
   LONG lStatus;
   HLINE hLine;
   HCALL hCall;
   DWORD dwVersion, dwCount;
   LINEEXTENSIONID led;
   LPLINEDEVCAPS pldc;
   LINECALLPARAMS lcp;
   BYTE *bBuffer;
   int iNum;
   char *szStr, *szPtr, *lpszOffset;
   VARSTRING *vs;


   if( ! dwNum )        //  No devices? Don't even dignify that with an attempt
      return( FALSE );

   bBuffer = GlobalAllocPtr( GHND, 4096 );       //  Allocate some buffer space
   if( ! bBuffer )
      return( FALSE );

   szStr = (LPSTR)bBuffer + 1024;                  //  Split into other buffers
   vs = (VARSTRING *)( szStr + 2048 );
   *szStr = 0;
   szPtr = szStr;

   for( dwCount = 0; dwCount < dwNum; dwCount++ )         //  Check each device
   {
      pldc = (LPLINEDEVCAPS)bBuffer;           //  Get an API version value and
      pldc->dwTotalSize = 1024;
      lStatus = lineNegotiateAPIVersion( hLineApp, dwCount, 
                                         0x00010003, 0x00010003,
                                         &dwVersion, &led );
      if( ! lStatus )                      //  if the version is valid, get the
      {                                  //  capabilities of the current device
         pldc->dwTotalSize = 1024;
         lStatus = lineGetDevCaps( hLineApp, dwCount, dwVersion, 0,
                                   pldc );
         if( pldc->dwMediaModes & LINEMEDIAMODE_DATAMODEM )     //  Data modem?
         {                                //  Add the modem name to a string to
            lpszOffset = (LPSTR)pldc + pldc->dwProviderInfoOffset;  //  let the
            lpszOffset = lpszOffset + lstrlen( lpszOffset ) + 1;  //  user pick
            lstrcpy( szPtr, lpszOffset );                //  an available modem

            szPtr = szPtr + lstrlen( lpszOffset ) + 1;
         }
      }
   }
   if( szPtr == szStr )                       //  No modems found? Quit if true
   {
      GlobalFreePtr( bBuffer );
      return( FALSE );
   }
   *( szPtr + 1 ) = '\0';          //  End modem name string with a double NULL
                          //  Display modems found and get a selection (plus 1)
   iNum = DialogBoxParam( hInst, MAKEINTRESOURCE( 10002 ), hWnd, ModemDlgProc,
                          (LPARAM)szStr );

   if( ! iNum )            //  Don't continue if CANCEL pressed in modem dialog
   {
      GlobalFreePtr( bBuffer );
      return( FALSE );
   }

   hLine = NULL;               //  Attempt to open a line to the selected modem
   if( lineOpen( hLineApp, iNum - 1, &hLine, dwVersion, 0, (DWORD)hWnd,
                 LINECALLPRIVILEGE_OWNER | LINECALLPRIVILEGE_MONITOR,
                 LINEMEDIAMODE_DATAMODEM, NULL ) < 0 )
   {
      GlobalFreePtr( bBuffer );    //  Quit the routine if modem failed to open
      return( TRUE );
   }

   memset( &lcp, 0, sizeof(LINECALLPARAMS) );
   lcp.dwTotalSize = sizeof(LINECALLPARAMS);  //  Initialize structure required
   lcp.dwMinRate = 14400;                         //  to actually make the call
   lcp.dwMaxRate = 57600;
   lcp.dwMediaMode = LINEMEDIAMODE_DATAMODEM;
         
   hCall = NULL;                     //  Attempt the actual call from the modem
   lStatus = lineMakeCall( hLine, &hCall, lpszPhoneNumber, 0, &lcp );
   if( lStatus < 0 )
   {
      lineClose( hLine );           //  If dial failed, close the line and exit
      GlobalFreePtr( bBuffer );
      return( FALSE );
   }
   GlobalFreePtr( bBuffer );
   return( TRUE );                     //  If dial worked, quit in a good state
}                                                         //  End of OpenTAPI()

// ----------------------------------------------------------------------------
// === Displays available Modems and allows user selection ====================
// ----------------------------------------------------------------------------

BOOL WINAPI ModemDlgProc( HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam )
{
   switch( msg )
   {
      case WM_INITDIALOG:                  //  Upon starting this dialog box...
      {
         LPSTR lpStr;
      
         lpStr = (LPSTR)lParam;           //  lParam points to a set of strings
         while( *lpStr )                       //  such as "modem1\0modem2\0\0"
         {                        //  Add each modem to combo box for selection
            SendDlgItemMessage( hDlg, 100, CB_ADDSTRING, 0, (LPARAM)lpStr );
            lpStr = lpStr + lstrlen( lpStr ) + 1;
         }
         SendDlgItemMessage( hDlg, 100, CB_SETCURSEL, 0, 0 );   //  Set default
         return( TRUE );                               //  Wait for user action
      }


      case WM_COMMAND:
         if( wParam == IDOK )
         {
            int i;
                    //  If user selected OK button, get current combo selection
            i = SendDlgItemMessage( hDlg, 100, CB_GETCURSEL, 0, 0 );
            EndDialog( hDlg, i + 1 );    //  Return number of the selection + 1
            return( TRUE );                     //  so zero indicates a failure
         }
         if( wParam == IDCANCEL )
         {
            EndDialog( hDlg, FALSE );                 //  Return zero on cancel
            return( TRUE );
         }
         break;
   }
   return( FALSE );
}                                                     //  End of ModemDlgProc()

// ----------------------------------------------------------------------------
// ============================================================================
// ----------------------------------------------------------------------------

